@@ -22,38 +22,12 @@
 #include "condor_debug.h"
 #include "condor_attributes.h"
 #include "schedd_negotiate.h"
+#include "scheduler.h"
+#include "qmgmt.h"
 
 
-ResourceRequestCluster::ResourceRequestCluster(int auto_cluster_id):
-	m_auto_cluster_id(auto_cluster_id)
-{
-}
-
-void
-ResourceRequestCluster::addJob(PROC_ID job_id)
-{
-	m_job_ids.push_back( job_id );
-}
-
-bool
-ResourceRequestCluster::popJob(PROC_ID &job_id)
-{
-	if( m_job_ids.empty() ) {
-		return false;
-	}
-	job_id = m_job_ids.front();
-	m_job_ids.pop_front();
-	return true;
-}
-
-ResourceRequestList::~ResourceRequestList()
-{
-	while( !empty() ) {
-		ResourceRequestCluster *cluster = front();
-		pop_front();
-		delete cluster;
-	}
-}
+static int next_negotiate_instance_id = 1;
+extern Scheduler scheduler;
 
 ScheddNegotiate::ScheddNegotiate
 (
@@ -64,10 +38,6 @@ ScheddNegotiate::ScheddNegotiate
 ):
 	DCMsg(cmd),
 	m_jobs(jobs),
-	m_current_resources_requested(1),
-	m_current_resources_delivered(0),
-	m_jobs_can_offer(-1),
-	m_will_match_claimed_pslots(false),
 	m_owner(owner ? owner : ""),
 	m_remote_pool(remote_pool ? remote_pool : ""),
 	m_current_auto_cluster_id(-1),
@@ -81,11 +51,22 @@ ScheddNegotiate::ScheddNegotiate
 {
 	m_current_job_id.cluster = -1;
 	m_current_job_id.proc = -1;
+	m_instance = next_negotiate_instance_id++;
 }
 
 ScheddNegotiate::~ScheddNegotiate()
 {
 	delete m_jobs;
+	delete m_reject_ad; m_reject_ad = nullptr;
+}
+
+void
+ScheddNegotiate::setMatchCaps(std::string_view caps)
+{
+	for (auto & str : StringTokenIterator(caps)) {
+		if (YourStringNoCase("MatchDiag3") == str) { m_can_do_match_diag_3 = true; }
+		else if (YourStringNoCase("Dye") == str) { m_can_do_match_dye = true; }
+	}
 }
 
 void
@@ -97,6 +78,9 @@ ScheddNegotiate::negotiate(Sock *sock)
 		// consists of asynchronously reading messages from the
 		// negotiator and sending responses until this round of
 		// negotiation is done.
+	dprintf(D_DYE, "%08x_%06d:%s\tNEGOTIATE\t%s rr=%d\n",
+		sock->getUniqueId(), m_instance,
+		m_owner.c_str(), m_remote_pool.c_str(), m_jobs ? (int)m_jobs->size() : 0);
 
 		// This will eventually call ScheddNegotiate::readMsg().
 	classy_counted_ptr<DCMessenger> messenger = new DCMessenger(sock);
@@ -118,9 +102,45 @@ ScheddNegotiate::getRemotePool()
 	return m_remote_pool.c_str();
 }
 
+char const *
+ScheddNegotiate::getNegotiatorName()
+{
+	return m_negotiator_name.c_str();
+}
+
 // this is in qmgmt.cpp...
 extern JobQueueJob* GetJobAd(const PROC_ID& jid);
 
+// flatten/expand the items in the list so that each ResourceRequestCluster
+// holds only a single job.  We do this so that the old negotiator prototol
+// that doesn't understand resource request lists can be fed a single job at a time
+//
+void ResourceRequestList::flatten()
+{
+	if (largest_cluster <= 1) return;
+
+	std::deque<ResourceRequestCluster> tmp;
+	items.swap(tmp);
+
+	largest_cluster = 0;
+	for (auto & cluster : tmp) {
+		int id = cluster.getAutoClusterId();
+		for (auto & jid : cluster) {
+			items.emplace_back(id).addJob(jid);
+			largest_cluster = MAX(largest_cluster, items.back().size());
+		}
+	}
+}
+
+// when the negotiator sends a SEND_JOB_INFO instead of SEND_RESOURCE_REQUEST_LIST
+// then it doesn't want to know about resource request lists, so we re-write
+// the ResourceRequestList items so that each only holds a single job.
+void ScheddNegotiate::notSendingResourceRequests()
+{
+	if ( ! m_jobs || m_jobs->empty()) return;
+	m_jobs->flatten();
+}
+
 bool
 ScheddNegotiate::nextJob()
 {
@@ -138,18 +158,37 @@ ScheddNegotiate::nextJob()
 	}
 
 	while( !m_jobs->empty() && m_jobs_can_offer ) {
-		ResourceRequestCluster *cluster = m_jobs->front();
-		ASSERT( cluster );
+		ResourceRequestCluster & cluster = m_jobs->front();
 
-		m_current_auto_cluster_id = cluster->getAutoClusterId();
+		m_current_auto_cluster_id = cluster.getAutoClusterId();
 
+			// if this auto-cluster has already been rejected, skip it
 		if( !getAutoClusterRejected(m_current_auto_cluster_id) ) {
-			while( cluster->popJob(m_current_job_id) ) {
+			size_t clusterSize = cluster.size();
+			for (auto & jid : cluster) {
+				--clusterSize; // decrement as we iterate so we know how many jobs remain
+
+					// if this is not a job (i.e. an OCU request), skip it
+				if (!jid.isJobKey() && (jid.proc != OCU_qkey2)) {
+					continue;
+				}
+				m_current_job_id = jid;
+
 				const char* because = "";
 				bool skip_all = false;
 				JobQueueJob * job = GetJobAd(m_current_job_id);
-				if ( ! job)
-				{
+				if (! job) {
+					// Is this an OCU request?
+					if (scheduler.getOCU(m_current_job_id.cluster)) {
+						m_current_job_ad = scheduler.getOCU(m_current_job_id.cluster)->ad;
+						m_current_job_ad.Assign(ATTR_RESOURCE_REQUEST_COUNT, 1);
+
+						// negotiator has to see clusterid and procid or it will not be happy.  Sad but true.
+						m_current_job_ad.Assign(ATTR_CLUSTER_ID, m_current_job_id.cluster);
+						m_current_job_ad.Assign(ATTR_PROC_ID, OCU_qkey2);
+						m_current_job_ad.Assign(ATTR_REQUIREMENTS, true);
+						return true;
+					} 
 					dprintf(D_MATCH,
 						"skipping job %d.%d because it no longer exists\n",
 						m_current_job_id.cluster,m_current_job_id.proc);
@@ -187,8 +226,8 @@ ScheddNegotiate::nextJob()
 						m_current_job_ad.LookupInteger(ATTR_JOB_UNIVERSE,universe);
 						// For now, do not use request counts with the dedicated scheduler
 						if ( universe != CONDOR_UNIVERSE_PARALLEL ) {
-							// add one to cluster size to cover the current popped job
-							int resource_count = 1+cluster->size();
+							// resource_count is the remaining un-iterated jobs plus this one.
+							int resource_count = 1+clusterSize;
 							if (count_max > 0) { resource_count = MIN(resource_count, count_max); }
 							if (resource_count > m_jobs_can_offer && (m_jobs_can_offer > 0))
 							{
@@ -213,7 +252,6 @@ ScheddNegotiate::nextJob()
 		}
 
 		m_jobs->pop_front();
-		delete cluster;
 	}
 	if (!m_jobs_can_offer)
 	{
@@ -337,6 +375,7 @@ ScheddNegotiate::fixupPartitionableSlot(ClassAd *job_ad, ClassAd *match_ad)
 bool
 ScheddNegotiate::sendResourceRequestList(Sock *sock)
 {
+		// The Negotiator wants us to send it a list of resource requests.
 	m_jobs_can_offer = scheduler_maxJobsToOffer();
 
 	while (m_num_resource_reqs_to_send > 0) {
@@ -373,9 +412,7 @@ ScheddNegotiate::sendResourceRequestList(Sock *sock)
 		// we already sent an ad with a resource_request_count.  So we want
 		// to skip ahead to the next cluster.
 		if ( !m_jobs->empty() ) {
-			ResourceRequestCluster *cluster = m_jobs->front();
 			m_jobs->pop_front();
-			delete cluster;
 		}
 
 		m_num_resource_reqs_sent++;
@@ -414,6 +451,17 @@ ScheddNegotiate::sendJobInfo(Sock *sock, bool just_sig_attrs)
 		return false;
 	}
 
+	if (IsDebugCategory(D_DYE)) {
+		std::string details;
+		m_current_job_ad.LookupString(ATTR_OWNER, details);
+		details += " "; details += m_owner;
+		int count = 0, aid = 0;
+		m_current_job_ad.LookupInteger(ATTR_RESOURCE_REQUEST_COUNT, count);
+		m_current_job_ad.LookupInteger(ATTR_AUTO_CLUSTER_ID, aid);
+		formatstr_cat(details, " %d |%d|%d.%d|", count, aid, m_current_job_id.cluster, m_current_job_id.proc);
+		dprintf(D_DYE, "\t%s\n", details.c_str());
+	}
+
 		// Tell the negotiator that we understand pslot preemption
 	m_current_job_ad.Assign(ATTR_WANT_PSLOT_PREEMPTION, true);
 		
@@ -421,7 +469,11 @@ ScheddNegotiate::sendJobInfo(Sock *sock, bool just_sig_attrs)
 		// 0 = no match diagnostics
 		// 1 = match diagnostics string
 		// 2 = match diagnostics string decorated w/ autocluster + jobid
-	m_current_job_ad.Assign(ATTR_WANT_MATCH_DIAGNOSTICS, (int) 2);
+		// 3 = (same as 2) + optional diagnostics ad
+	int match_diag = 2;
+	// TODO: only turn on diag 3 only for jobs that request it?
+	if (m_can_do_match_diag_3) { match_diag = 3; }
+	m_current_job_ad.Assign(ATTR_WANT_MATCH_DIAGNOSTICS, (int) match_diag);
 
 		// Send the ad to the negotiator
 	int putad_result = 0;
@@ -498,6 +550,30 @@ ScheddNegotiate::messageReceived( DCMessenger *messenger, Sock *sock )
 		// This is called when readMsg() returns true.
 		// Now carry out the negotiator's request that we just read.
 
+	if (IsDebugCategory(D_DYE)) {
+		std::string details;
+		switch(m_operation) {
+		case REJECTED_WITH_REASON: details = m_reject_reason; break;
+		case SEND_JOB_INFO: {
+			if ( m_jobs && ! m_jobs->empty()) { auto & rr = m_jobs->front(); formatstr(details, "ac=%d", rr.getAutoClusterId()); }
+			} break;
+		case PERMISSION_AND_AD: {
+			m_match_ad.LookupString(ATTR_NAME,details);
+			JOB_ID_KEY jid;
+			m_match_ad.LookupInteger(ATTR_RESOURCE_REQUEST_CLUSTER, jid.cluster);
+			m_match_ad.LookupInteger(ATTR_RESOURCE_REQUEST_PROC, jid.proc);
+			formatstr_cat(details, " %d ", m_current_resources_delivered);
+			// TODO: have negotiator send RR id back, not just job id.
+			details += "|?|"; details += (std::string)(jid); details += "|";
+			} break;
+		case END_NEGOTIATE: {
+			if (RRLRequestIsPending()) { details = "<prefetch>"; } else { details = "<end>"; }
+			} break;
+		}
+		dprintf(D_DYE, "%08x_%06d:%s\t%s\t%s\n", sock->getUniqueId(), m_instance,
+			m_owner.c_str(), getCommandStringSafe(m_operation), details.c_str());
+	}
+
 	switch( m_operation ) {
 
 	case REJECTED:
@@ -528,9 +604,18 @@ ScheddNegotiate::messageReceived( DCMessenger *messenger, Sock *sock )
 				m_current_job_id.cluster = rr_cluster;
 				m_current_job_id.proc = rr_proc;
 			}
-			m_reject_reason.erase(pos);	// will truncate string at pos
+			m_reject_reason.erase(pos); // will truncate string at pos
+			trim(m_reject_reason);
 		}
-		scheduler_handleJobRejected( m_current_job_id, m_reject_reason.c_str() );
+		// did we get a rejection ad?
+		if (m_reject_ad && m_reject_ad->size() > 0) {
+			if (IsDebugVerbose(D_MATCH)) {
+				std::string adbuf;
+				dprintf(D_MATCH | D_VERBOSE, "Got a match rejection ad from %s:\n%s",
+					m_negotiator_name.c_str(), formatAd(adbuf, *m_reject_ad, "\t"));
+			}
+		}
+		scheduler_handleJobRejected( m_current_job_id, m_current_auto_cluster_id, m_reject_reason.c_str() );
 		m_jobs_rejected++;
 		setAutoClusterRejected( m_current_auto_cluster_id );
 		nextJob();
@@ -540,6 +625,7 @@ ScheddNegotiate::messageReceived( DCMessenger *messenger, Sock *sock )
 	case SEND_JOB_INFO:
 		//dprintf(D_MATCH, "got SEND_JOB_INFO\n");
 		m_num_resource_reqs_sent = 0;  // clear counter of reqs sent this round
+		notSendingResourceRequests();
 		if( !sendJobInfo(sock) ) {
 				// We failed to talk to the negotiator, so close the socket.
 			return MESSAGE_FINISHED;
@@ -568,6 +654,7 @@ ScheddNegotiate::messageReceived( DCMessenger *messenger, Sock *sock )
 		int rr_proc = -1;
 		m_match_ad.LookupInteger(ATTR_RESOURCE_REQUEST_CLUSTER, rr_cluster);
 		m_match_ad.LookupInteger(ATTR_RESOURCE_REQUEST_PROC, rr_proc);
+			// if we got a valid job id back, use it to update m_current_job_id
 		if (rr_cluster != -1 && rr_proc != -1) {
 			if (rr_cluster != m_current_job_id.cluster || rr_proc != m_current_job_id.proc) {
 				m_current_resources_delivered = 0;
@@ -592,7 +679,8 @@ ScheddNegotiate::messageReceived( DCMessenger *messenger, Sock *sock )
 			break;
 		}
 
-		if( scheduler_handleMatch(m_current_job_id,m_claim_id.c_str(),m_extra_claims.c_str(), m_match_ad,slot_name) )
+		auto normal = ScheddNegotiate::_match_source::NORMAL;
+		if( scheduler_handleMatch(m_current_job_id,m_claim_id.c_str(),m_extra_claims.c_str(), m_match_ad, slot_name, normal) )
 		{
 			m_jobs_matched++;
 		}
@@ -604,7 +692,7 @@ ScheddNegotiate::messageReceived( DCMessenger *messenger, Sock *sock )
 
 	case END_NEGOTIATE:
 		if (RRLRequestIsPending()) {
-			dprintf(D_ALWAYS, "Finished sending rrls to negotiator\n");
+			dprintf(D_FULLDEBUG, "Finished RRL prefetch to negotiator\n");
 		} else {
 			dprintf( D_ALWAYS, "Negotiation ended: %d jobs matched\n",
 					 m_jobs_matched );
@@ -657,14 +745,24 @@ ScheddNegotiate::readMsg( DCMessenger * /*messenger*/, Sock *sock )
 
 	switch( m_operation ) {
 	case REJECTED_WITH_REASON:
+		if (m_reject_ad) { m_reject_ad->Clear(); } // in case we don't get a rejection ad
 		if( !sock->code(m_reject_reason) ) {
 			dprintf( D_ALWAYS,
 					 "Can't receive reject reason from negotiator\n" );
 			return false;
 		}
+		// a 24.X negotiator may also return a match diagnostics ad
+		if (m_can_do_match_diag_3 && ! sock->peek_end_of_message()) {
+			if ( ! m_reject_ad) { m_reject_ad = new ClassAd(); }
+			if ( ! getClassAd(sock, *m_reject_ad)) {
+				dprintf( D_ALWAYS, "Can't get my reject ad from negotiator\n" );
+				return false;
+			}
+		}
 		break;
 
 	case REJECTED:
+		if (m_reject_ad) { m_reject_ad->Clear(); } // in case we don't get a rejection ad
 		break;
 
 	case SEND_JOB_INFO:
