@@ -19,7 +19,7 @@ import sys
 import logging
 import argparse
 
-import htcondor
+import htcondor2 as htcondor
 
 from pathlib import Path
 
@@ -38,11 +38,15 @@ def get_default_config(name="ADSTASH"):
         "collectors": htcondor.param.get("CONDOR_HOST"),
         "read_schedd_history": False,
         "read_startd_history": False,
+        "read_schedd_job_epoch_history": False,
+        "read_schedd_transfer_epoch_history": False,
         "read_ad_file": None,
         "schedd_history_max_ads": 10000,
         "startd_history_max_ads": 10000,
         "schedd_history_timeout": 2 * 60,
         "startd_history_timeout": 2 * 60,
+        "schedd_history_projection": set(),
+        "startd_history_projection": set(),
         "interface": "null",
         "se_host": "localhost:9200",
         "se_use_https": False,
@@ -70,11 +74,15 @@ def get_htcondor_config(name="ADSTASH"):
         "startds": p.get(f"{name}_READ_STARTDS"),
         "read_schedd_history": p.get(f"{name}_SCHEDD_HISTORY"),
         "read_startd_history": p.get(f"{name}_STARTD_HISTORY"),
+        "read_schedd_job_epoch_history": p.get(f"{name}_SCHEDD_JOB_EPOCH_HISTORY"),
+        "read_schedd_transfer_epoch_history": p.get(f"{name}_SCHEDD_TRANSFER_EPOCH_HISTORY"),
         "read_ad_file": p.get(f"{name}_AD_FILE"),
         "schedd_history_max_ads": p.get(f"{name}_SCHEDD_HISTORY_MAX_ADS"),
         "startd_history_max_ads": p.get(f"{name}_STARTD_HISTORY_MAX_ADS"),
         "schedd_history_timeout": p.get(f"{name}_SCHEDD_HISTORY_TIMEOUT"),
         "startd_history_timeout": p.get(f"{name}_STARTD_HISTORY_TIMEOUT"),
+        "schedd_history_projection": p.get(f"{name}_SCHEDD_HISTORY_PROJECTION"),
+        "startd_history_projection": p.get(f"{name}_STARTD_HISTORY_PROJECTION"),
         "interface": p.get(f"{name}_INTERFACE"),
         "se_host": p.get(f"{name}_ES_HOST", p.get(f"{name}_SE_HOST")),
         "se_url_prefix": p.get(f"{name}_SE_URL_PREFIX"),
@@ -135,11 +143,15 @@ def get_environment_config(name="ADSTASH"):
         "startds": env.get(f"{name}_READ_STARTDS"),
         "read_schedd_history": env.get(f"{name}_SCHEDD_HISTORY"),
         "read_startd_history": env.get(f"{name}_STARTD_HISTORY"),
+        "read_schedd_job_epoch_history": env.get(f"{name}_SCHEDD_JOB_EPOCH_HISTORY"),
+        "read_schedd_transfer_epoch_history": env.get(f"{name}_SCHEDD_TRANSFER_EPOCH_HISTORY"),
         "read_ad_file": env.get(f"{name}_AD_FILE"),
         "schedd_history_max_ads": env.get(f"{name}_SCHEDD_HISTORY_MAX_ADS"),
         "startd_history_max_ads": env.get(f"{name}_STARTD_HISTORY_MAX_ADS"),
         "schedd_history_timeout": env.get(f"{name}_SCHEDD_HISTORY_TIMEOUT"),
         "startd_history_timeout": env.get(f"{name}_STARTD_HISTORY_TIMEOUT"),
+        "schedd_history_projection": env.get(f"{name}_SCHEDD_HISTORY_PROJECTION"),
+        "startd_history_projection": env.get(f"{name}_STARTD_HISTORY_PROJECTION"),
         "interface": env.get(f"{name}_INTERFACE"),
         "se_host": env.get(f"{name}_SE_HOST", env.get(f"{name}_ES_HOST")),
         "se_url_prefix": env.get(f"{name}_SE_URL_PREFIX"),
@@ -221,6 +233,8 @@ def normalize_config_types(conf):
         "standalone",
         "read_schedd_history",
         "read_startd_history",
+        "read_schedd_job_epoch_history",
+        "read_schedd_transfer_epoch_history",
         "to_elasticsearch",
         "to_json",
         "se_use_https",
@@ -381,6 +395,24 @@ def get_config(argv=None):
             "[default: %(default)s]"
         ),
     )
+    source_group.add_argument(
+        "--schedd_job_epoch_history",
+        action="store_true",
+        dest="read_schedd_job_epoch_history",
+        help=(
+            "Poll Schedd Job Epoch histories "
+            "[default: %(default)s]"
+        ),
+    )
+    source_group.add_argument(
+        "--schedd_transfer_epoch_history",
+        action="store_true",
+        dest="read_schedd_transfer_epoch_history",
+        help=(
+            "Poll Schedd Transfer Epoch histories "
+            "[default: %(default)s]"
+        ),
+    )
     source_group.add_argument(
         "--ad_file",
         type=Path,
@@ -450,6 +482,24 @@ def get_config(argv=None):
             "[default: %(default)s]"
         ),
     )
+    history_group.add_argument(
+        "--schedd_history_projection",
+        type=str,
+        metavar="ATTRIBUTES",
+        help=(
+            "Fetch only these attributes (comma-separated) from the Schedd history "
+            "[default: fetch all attributes]"
+        ),
+    )
+    history_group.add_argument(
+        "--startd_history_projection",
+        type=str,
+        metavar="ATTRIBUTES",
+        help=(
+            "Fetch only these attributes (comma-separated) from the Startd history "
+            "[default: fetch all attributes]"
+        ),
+    )
 
     se_group = parser.add_argument_group(
         title = "Search engine (Elasticsearch, OpenSearch, etc.) interface options",
@@ -535,6 +585,11 @@ def get_config(argv=None):
     args_dict = vars(args)
     args_dict['process_name'] = process_name
 
+    if args.schedd_history_projection:
+        args.schedd_history_projection = set(re.split(r"[\s,]+", args.schedd_history_projection.strip()))
+    if args.startd_history_projection:
+        args.startd_history_projection = set(re.split(r"[\s,]+", args.startd_history_projection.strip()))
+
     # Check for deprecated args
     for arg in remaining_argv:
         if arg.startswith("--es_"):
